@@ -14,6 +14,7 @@
 #include <fcntl.h>
 #include <pwd.h>
 #include <grp.h>
+#include <blkid.h>
 
 #include "strutils.h"
 #include "pathnames.h"
@@ -22,40 +23,37 @@
 #include "canonicalize.h"
 #include "env.h"
 #include "match.h"
+#include "fileutils.h"
+#include "statfs_magic.h"
+#include "sysfs.h"
 
-int endswith(const char *s, const char *sx)
+int append_string(char **a, const char *b)
 {
-	ssize_t off;
+	size_t al, bl;
+	char *tmp;
 
-	assert(s);
-	assert(sx);
+	assert(a);
 
-	off = strlen(s);
-	if (!off)
-		return 0;
-	off -= strlen(sx);
-	if (off < 0)
+	if (!b || !*b)
 		return 0;
+	if (!*a) {
+		*a = strdup(b);
+		return !*a ? -ENOMEM : 0;
+	}
 
-        return !strcmp(s + off, sx);
-}
+	al = strlen(*a);
+	bl = strlen(b);
 
-int startswith(const char *s, const char *sx)
-{
-	size_t off;
-
-	assert(s);
-	assert(sx);
-
-	off = strlen(sx);
-	if (!off)
-		return 0;
-
-        return !strncmp(s, sx, off);
+	tmp = realloc(*a, al + bl + 1);
+	if (!tmp)
+		return -ENOMEM;
+	*a = tmp;
+	memcpy((*a) + al, b, bl + 1);
+	return 0;
 }
 
 /*
- * Return 1 if the file does not accessible of empty
+ * Return 1 if the file is not accessible or empty
  */
 int is_file_empty(const char *name)
 {
@@ -65,6 +63,34 @@ int is_file_empty(const char *name)
 	return (stat(name, &st) != 0 || st.st_size == 0);
 }
 
+int mnt_valid_tagname(const char *tagname)
+{
+	if (tagname && *tagname && (
+	    strcmp("UUID", tagname) == 0 ||
+	    strcmp("LABEL", tagname) == 0 ||
+	    strcmp("PARTUUID", tagname) == 0 ||
+	    strcmp("PARTLABEL", tagname) == 0))
+		return 1;
+
+	return 0;
+}
+
+/**
+ * mnt_tag_is_valid:
+ * @tag: NAME=value string
+ *
+ * Returns: 1 if the @tag is parsable and tag NAME= is supported by libmount, or 0.
+ */
+int mnt_tag_is_valid(const char *tag)
+{
+	char *t = NULL;
+	int rc = tag && blkid_parse_tag_string(tag, &t, NULL) == 0
+		     && mnt_valid_tagname(t);
+
+	free(t);
+	return rc;
+}
+
 int mnt_parse_offset(const char *str, size_t len, uintmax_t *res)
 {
 	char *p;
@@ -92,21 +118,18 @@ static int fstype_cmp(const void *v1, const void *v2)
 	return strcmp(s1, s2);
 }
 
-/* returns basename and keeps dirname in the @path, if @path is "/" (root)
- * then returns empty string */
-char *stripoff_last_component(char *path)
+int mnt_stat_mountpoint(const char *target, struct stat *st)
 {
-	char *p = path ? strrchr(path, '/') : NULL;
-
-	if (!p)
-		return NULL;
-	*p = '\0';
-	return p + 1;
+#ifdef AT_NO_AUTOMOUNT
+	return fstatat(-1, target, st, AT_NO_AUTOMOUNT);
+#else
+	return stat(target, st);
+#endif
 }
 
 /*
- * Note that the @target has to be absolute path (so at least "/").  The
- * @filename returns allocated buffer with last path component, for example:
+ * Note that the @target has to be an absolute path (so at least "/").  The
+ * @filename returns an allocated buffer with the last path component, for example:
  *
  * mnt_chdir_to_parent("/mnt/test", &buf) ==> chdir("/mnt"), buf="test"
  */
@@ -119,7 +142,7 @@ int mnt_chdir_to_parent(const char *target, char **filename)
 	if (!target || *target != '/')
 		return -EINVAL;
 
-	DBG(UTILS, mnt_debug("moving to %s parent", target));
+	DBG(UTILS, ul_debug("moving to %s parent", target));
 
 	buf = strdup(target);
 	if (!buf)
@@ -134,22 +157,22 @@ int mnt_chdir_to_parent(const char *target, char **filename)
 	parent = buf && *buf ? buf : "/";
 
 	if (chdir(parent) == -1) {
-		DBG(UTILS, mnt_debug("failed to chdir to %s: %m", parent));
+		DBG(UTILS, ul_debug("failed to chdir to %s: %m", parent));
 		rc = -errno;
 		goto err;
 	}
 	if (!getcwd(cwd, sizeof(cwd))) {
-		DBG(UTILS, mnt_debug("failed to obtain current directory: %m"));
+		DBG(UTILS, ul_debug("failed to obtain current directory: %m"));
 		rc = -errno;
 		goto err;
 	}
 	if (strcmp(cwd, parent) != 0) {
-		DBG(UTILS, mnt_debug(
+		DBG(UTILS, ul_debug(
 		    "unexpected chdir (expected=%s, cwd=%s)", parent, cwd));
 		goto err;
 	}
 
-	DBG(CXT, mnt_debug(
+	DBG(CXT, ul_debug(
 		"current directory moved to %s [last_component='%s']",
 		parent, last));
 
@@ -159,7 +182,7 @@ int mnt_chdir_to_parent(const char *target, char **filename)
 		if (!last || !*last)
 			memcpy(*filename, ".", 2);
 		else
-			memcpy(*filename, last, strlen(last) + 1);
+			memmove(*filename, last, strlen(last) + 1);
 	} else
 		free(buf);
 	return 0;
@@ -169,7 +192,7 @@ err:
 }
 
 /*
- * Check if @path is on read-only filesystem independently on file permissions.
+ * Check if @path is on a read-only filesystem independently of file permissions.
  */
 int mnt_is_readonly(const char *path)
 {
@@ -180,7 +203,7 @@ int mnt_is_readonly(const char *path)
 	if (errno != EACCES)
 		return 0;
 
-#ifdef HAVE_FUTIMENS
+#ifdef HAVE_UTIMENSAT
 	/*
 	 * access(2) returns EACCES on read-only FS:
 	 *
@@ -188,7 +211,7 @@ int mnt_is_readonly(const char *path)
 	 *   accessible for the current rUID. (Note that euidaccess(2) does not
 	 *   check for EROFS at all).
 	 *
-	 * - for read-write filesystem with read-only VFS node (aka -o remount,ro,bind)
+	 * - for a read-write filesystem with a read-only VFS node (aka -o remount,ro,bind)
 	 */
 	{
 		struct timespec times[2];
@@ -209,7 +232,7 @@ int mnt_is_readonly(const char *path)
  *
  * Encode @str to be compatible with fstab/mtab
  *
- * Returns: new allocated string or NULL in case of error.
+ * Returns: newly allocated string or NULL in case of error.
  */
 char *mnt_mangle(const char *str)
 {
@@ -222,7 +245,7 @@ char *mnt_mangle(const char *str)
  *
  * Decode @str from fstab/mtab
  *
- * Returns: new allocated string or NULL in case of error.
+ * Returns: newly allocated string or NULL in case of error.
  */
 char *mnt_unmangle(const char *str)
 {
@@ -244,6 +267,7 @@ int mnt_fstype_is_pseudofs(const char *type)
 		"bdev",
 		"binfmt_misc",
 		"cgroup",
+		"cgroup2",
 		"configfs",
 		"cpuset",
 		"debugfs",
@@ -251,12 +275,14 @@ int mnt_fstype_is_pseudofs(const char *type)
 		"devpts",
 		"devtmpfs",
 		"dlmfs",
+		"efivarfs",
 		"fuse.gvfs-fuse-daemon",
 		"fusectl",
 		"hugetlbfs",
 		"mqueue",
 		"nfsd",
 		"none",
+		"overlay",
 		"pipefs",
 		"proc",
 		"pstore",
@@ -284,8 +310,6 @@ int mnt_fstype_is_pseudofs(const char *type)
  */
 int mnt_fstype_is_netfs(const char *type)
 {
-	assert(type);
-
 	if (strcmp(type, "cifs")   == 0 ||
 	    strcmp(type, "smbfs")  == 0 ||
 	    strncmp(type,"nfs", 3) == 0 ||
@@ -296,12 +320,98 @@ int mnt_fstype_is_netfs(const char *type)
 	return 0;
 }
 
+const char *mnt_statfs_get_fstype(struct statfs *vfs)
+{
+	assert(vfs);
+
+	switch (vfs->f_type) {
+	case STATFS_ADFS_MAGIC:		return "adfs";
+	case STATFS_AFFS_MAGIC:		return "affs";
+	case STATFS_AFS_MAGIC:		return "afs";
+	case STATFS_AUTOFS_MAGIC:	return "autofs";
+	case STATFS_BDEVFS_MAGIC:	return "bdev";
+	case STATFS_BEFS_MAGIC:		return "befs";
+	case STATFS_BFS_MAGIC:		return "befs";
+	case STATFS_BINFMTFS_MAGIC:	return "binfmt_misc";
+	case STATFS_BTRFS_MAGIC:	return "btrfs";
+	case STATFS_CEPH_MAGIC:		return "ceph";
+	case STATFS_CGROUP_MAGIC:	return "cgroup";
+	case STATFS_CIFS_MAGIC:		return "cifs";
+	case STATFS_CODA_MAGIC:		return "coda";
+	case STATFS_CONFIGFS_MAGIC:	return "configfs";
+	case STATFS_CRAMFS_MAGIC:	return "cramfs";
+	case STATFS_DEBUGFS_MAGIC:	return "debugfs";
+	case STATFS_DEVPTS_MAGIC:	return "devpts";
+	case STATFS_ECRYPTFS_MAGIC:	return "ecryptfs";
+	case STATFS_EFIVARFS_MAGIC:	return "efivarfs";
+	case STATFS_EFS_MAGIC:		return "efs";
+	case STATFS_EXOFS_MAGIC:	return "exofs";
+	case STATFS_EXT4_MAGIC:		return "ext4";	   /* all extN use the same magic */
+	case STATFS_F2FS_MAGIC:		return "f2fs";
+	case STATFS_FUSE_MAGIC:		return "fuse";
+	case STATFS_FUTEXFS_MAGIC:	return "futexfs";
+	case STATFS_GFS2_MAGIC:		return "gfs2";
+	case STATFS_HFSPLUS_MAGIC:	return "hfsplus";
+	case STATFS_HOSTFS_MAGIC:	return "hostfs";
+	case STATFS_HPFS_MAGIC:		return "hpfs";
+	case STATFS_HPPFS_MAGIC:	return "hppfs";
+	case STATFS_HUGETLBFS_MAGIC:	return "hugetlbfs";
+	case STATFS_ISOFS_MAGIC:	return "iso9660";
+	case STATFS_JFFS2_MAGIC:	return "jffs2";
+	case STATFS_JFS_MAGIC:		return "jfs";
+	case STATFS_LOGFS_MAGIC:	return "logfs";
+	case STATFS_MINIX2_MAGIC:
+	case STATFS_MINIX2_MAGIC2:
+	case STATFS_MINIX3_MAGIC:
+	case STATFS_MINIX_MAGIC:
+	case STATFS_MINIX_MAGIC2:	return "minix";
+	case STATFS_MQUEUE_MAGIC:	return "mqueue";
+	case STATFS_MSDOS_MAGIC:	return "vfat";
+	case STATFS_NCP_MAGIC:		return "ncp";
+	case STATFS_NFS_MAGIC:		return "nfs";
+	case STATFS_NILFS_MAGIC:	return "nilfs2";
+	case STATFS_NTFS_MAGIC:		return "ntfs";
+	case STATFS_OCFS2_MAGIC:	return "ocfs2";
+	case STATFS_OMFS_MAGIC:		return "omfs";
+	case STATFS_OPENPROMFS_MAGIC:	return "openpromfs";
+	case STATFS_PIPEFS_MAGIC:	return "pipefs";
+	case STATFS_PROC_MAGIC:		return "proc";
+	case STATFS_PSTOREFS_MAGIC:	return "pstore";
+	case STATFS_QNX4_MAGIC:		return "qnx4";
+	case STATFS_QNX6_MAGIC:		return "qnx6";
+	case STATFS_RAMFS_MAGIC:	return "ramfs";
+	case STATFS_REISERFS_MAGIC:	return "reiser4";
+	case STATFS_ROMFS_MAGIC:	return "romfs";
+	case STATFS_SECURITYFS_MAGIC:	return "securityfs";
+	case STATFS_SELINUXFS_MAGIC:	return "selinuxfs";
+	case STATFS_SMACKFS_MAGIC:	return "smackfs";
+	case STATFS_SMB_MAGIC:		return "smb";
+	case STATFS_SOCKFS_MAGIC:	return "sockfs";
+	case STATFS_SQUASHFS_MAGIC:	return "squashfs";
+	case STATFS_SYSFS_MAGIC:	return "sysfs";
+	case STATFS_TMPFS_MAGIC:	return "tmpfs";
+	case STATFS_UBIFS_MAGIC:	return "ubifs";
+	case STATFS_UDF_MAGIC:		return "udf";
+	case STATFS_UFS2_MAGIC:
+	case STATFS_UFS_MAGIC:		return "ufs";
+	case STATFS_V9FS_MAGIC:		return "9p";
+	case STATFS_VXFS_MAGIC:		return "vxfs";
+	case STATFS_XENFS_MAGIC:	return "xenfs";
+	case STATFS_XFS_MAGIC:		return "xfs";
+	default:
+		break;
+	}
+
+	return NULL;
+}
+
+
 /**
  * mnt_match_fstype:
  * @type: filesystem type
  * @pattern: filesystem name or comma delimited list of names
  *
- * The @pattern list of filesystem can be prefixed with a global
+ * The @pattern list of filesystems can be prefixed with a global
  * "no" prefix to invert matching of the whole list. The "no" could
  * also be used for individual items in the @pattern list. So,
  * "nofoo,bar" has the same meaning as "nofoo,nobar".
@@ -359,14 +469,14 @@ static int check_option(const char *haystack, size_t len,
  * @optstr: options string
  * @pattern: comma delimited list of options
  *
- * The "no" could used for individual items in the @options list. The "no"
+ * The "no" could be used for individual items in the @options list. The "no"
  * prefix does not have a global meaning.
  *
  * Unlike fs type matching, nonetdev,user and nonetdev,nouser have
  * DIFFERENT meanings; each option is matched explicitly as specified.
  *
- * The "no" prefix interpretation could be disable by "+" prefix, for example
- * "+noauto" matches if @optstr literally contains "noauto" string.
+ * The "no" prefix interpretation could be disabled by the "+" prefix, for example
+ * "+noauto" matches if @optstr literally contains the "noauto" string.
  *
  * "xxx,yyy,zzz" : "nozzz"	-> False
  *
@@ -451,10 +561,10 @@ static int add_filesystem(char ***filesystems, char *name)
 		*filesystems = x;
 	}
 	name = strdup(name);
-	if (!name)
-		goto err;
 	(*filesystems)[n] = name;
 	(*filesystems)[n + 1] = NULL;
+	if (!name)
+		goto err;
 	return 0;
 err:
 	mnt_free_filesystems(*filesystems);
@@ -465,13 +575,13 @@ static int get_filesystems(const char *filename, char ***filesystems, const char
 {
 	int rc = 0;
 	FILE *f;
-	char line[128];
+	char line[129];
 
 	f = fopen(filename, "r" UL_CLOEXECSTR);
 	if (!f)
 		return 1;
 
-	DBG(UTILS, mnt_debug("reading filesystems list from: %s", filename));
+	DBG(UTILS, ul_debug("reading filesystems list from: %s", filename));
 
 	while (fgets(line, sizeof(line), f)) {
 		char name[sizeof(line)];
@@ -496,7 +606,7 @@ static int get_filesystems(const char *filename, char ***filesystems, const char
 }
 
 /*
- * Always check @filesystems pointer!
+ * Always check the @filesystems pointer!
  *
  * man mount:
  *
@@ -521,36 +631,25 @@ int mnt_get_filesystems(char ***filesystems, const char *pattern)
 
 	rc = get_filesystems(_PATH_PROC_FILESYSTEMS, filesystems, pattern);
 	if (rc == 1 && *filesystems)
-		rc = 0;			/* not found /proc/filesystems */
+		rc = 0;			/* /proc/filesystems not found */
 
 	return rc;
 }
 
-static size_t get_pw_record_size(void)
-{
-#ifdef _SC_GETPW_R_SIZE_MAX
-	long sz = sysconf(_SC_GETPW_R_SIZE_MAX);
-	if (sz > 0)
-		return sz;
-#endif
-	return 16384;
-}
-
 /*
- * Returns allocated string with username or NULL.
+ * Returns an allocated string with username or NULL.
  */
 char *mnt_get_username(const uid_t uid)
 {
         struct passwd pwd;
 	struct passwd *res;
-	size_t sz = get_pw_record_size();
 	char *buf, *username = NULL;
 
-	buf = malloc(sz);
+	buf = malloc(UL_GETPW_BUFSIZ);
 	if (!buf)
 		return NULL;
 
-	if (!getpwuid_r(uid, &pwd, buf, sz, &res) && res)
+	if (!getpwuid_r(uid, &pwd, buf, UL_GETPW_BUFSIZ, &res) && res)
 		username = strdup(pwd.pw_name);
 
 	free(buf);
@@ -562,21 +661,20 @@ int mnt_get_uid(const char *username, uid_t *uid)
 	int rc = -1;
         struct passwd pwd;
 	struct passwd *pw;
-	size_t sz = get_pw_record_size();
 	char *buf;
 
 	if (!username || !uid)
 		return -EINVAL;
 
-	buf = malloc(sz);
+	buf = malloc(UL_GETPW_BUFSIZ);
 	if (!buf)
 		return -ENOMEM;
 
-	if (!getpwnam_r(username, &pwd, buf, sz, &pw) && pw) {
+	if (!getpwnam_r(username, &pwd, buf, UL_GETPW_BUFSIZ, &pw) && pw) {
 		*uid= pw->pw_uid;
 		rc = 0;
 	} else {
-		DBG(UTILS, mnt_debug(
+		DBG(UTILS, ul_debug(
 			"cannot convert '%s' username to UID", username));
 		rc = errno ? -errno : -EINVAL;
 	}
@@ -590,21 +688,20 @@ int mnt_get_gid(const char *groupname, gid_t *gid)
 	int rc = -1;
         struct group grp;
 	struct group *gr;
-	size_t sz = get_pw_record_size();
 	char *buf;
 
 	if (!groupname || !gid)
 		return -EINVAL;
 
-	buf = malloc(sz);
+	buf = malloc(UL_GETPW_BUFSIZ);
 	if (!buf)
 		return -ENOMEM;
 
-	if (!getgrnam_r(groupname, &grp, buf, sz, &gr) && gr) {
+	if (!getgrnam_r(groupname, &grp, buf, UL_GETPW_BUFSIZ, &gr) && gr) {
 		*gid= gr->gr_gid;
 		rc = 0;
 	} else {
-		DBG(UTILS, mnt_debug(
+		DBG(UTILS, ul_debug(
 			"cannot convert '%s' groupname to GID", groupname));
 		rc = errno ? -errno : -EINVAL;
 	}
@@ -663,8 +760,8 @@ static int try_write(const char *filename)
  * @mtab: returns path to mtab
  * @writable: returns 1 if the file is writable
  *
- * If the file does not exist and @writable argument is not NULL then it will
- * try to create the file
+ * If the file does not exist and @writable argument is not NULL, then it will
+ * try to create the file.
  *
  * Returns: 1 if /etc/mtab is a regular file, and 0 in case of error (check
  *          errno for more details).
@@ -680,15 +777,16 @@ int mnt_has_regular_mtab(const char **mtab, int *writable)
 	if (mtab && !*mtab)
 		*mtab = filename;
 
-	DBG(UTILS, mnt_debug("mtab: %s", filename));
+	DBG(UTILS, ul_debug("mtab: %s", filename));
 
 	rc = lstat(filename, &st);
 
 	if (rc == 0) {
-		/* file exist */
+		/* file exists */
 		if (S_ISREG(st.st_mode)) {
 			if (writable)
 				*writable = !try_write(filename);
+			DBG(UTILS, ul_debug("%s: writable", filename));
 			return 1;
 		}
 		goto done;
@@ -697,19 +795,21 @@ int mnt_has_regular_mtab(const char **mtab, int *writable)
 	/* try to create the file */
 	if (writable) {
 		*writable = !try_write(filename);
-		if (*writable)
+		if (*writable) {
+			DBG(UTILS, ul_debug("%s: writable", filename));
 			return 1;
+		}
 	}
 
 done:
-	DBG(UTILS, mnt_debug("%s: irregular/non-writable", filename));
+	DBG(UTILS, ul_debug("%s: irregular/non-writable", filename));
 	return 0;
 }
 
 /*
  * Don't export this to libmount API -- utab is private library stuff.
  *
- * If the file does not exist and @writable argument is not NULL then it will
+ * If the file does not exist and @writable argument is not NULL, then it will
  * try to create the directory (e.g. /run/mount) and the file.
  *
  * Returns: 1 if utab is a regular file, and 0 in case of
@@ -726,18 +826,18 @@ int mnt_has_regular_utab(const char **utab, int *writable)
 	if (utab && !*utab)
 		*utab = filename;
 
-	DBG(UTILS, mnt_debug("utab: %s", filename));
+	DBG(UTILS, ul_debug("utab: %s", filename));
 
 	rc = lstat(filename, &st);
 
 	if (rc == 0) {
-		/* file exist */
+		/* file exists */
 		if (S_ISREG(st.st_mode)) {
 			if (writable)
 				*writable = !try_write(filename);
 			return 1;
 		}
-		goto done;	/* it's not regular file */
+		goto done;	/* it's not a regular file */
 	}
 
 	if (writable) {
@@ -760,7 +860,7 @@ int mnt_has_regular_utab(const char **utab, int *writable)
 			return 1;
 	}
 done:
-	DBG(UTILS, mnt_debug("%s: irregular/non-writable file", filename));
+	DBG(UTILS, ul_debug("%s: irregular/non-writable file", filename));
 	return 0;
 }
 
@@ -789,7 +889,7 @@ const char *mnt_get_fstab_path(void)
 /**
  * mnt_get_mtab_path:
  *
- * This function returns *default* location of the mtab file. The result does
+ * This function returns the *default* location of the mtab file. The result does
  * not have to be writable. See also mnt_has_regular_mtab().
  *
  * Returns: path to /etc/mtab or $LIBMOUNT_MTAB.
@@ -820,7 +920,7 @@ const char *mnt_get_utab_path(void)
 }
 
 
-/* returns file descriptor or -errno, @name returns uniques filename
+/* returns file descriptor or -errno, @name returns a unique filename
  */
 int mnt_open_uniq_filename(const char *filename, char **name)
 {
@@ -837,12 +937,14 @@ int mnt_open_uniq_filename(const char *filename, char **name)
 	if (rc <= 0)
 		return -errno;
 
-	/* This is for very old glibc and for compatibility with Posix where is
+	/* This is for very old glibc and for compatibility with Posix, which says
 	 * nothing about mkstemp() mode. All sane glibc use secure mode (0600).
 	 */
 	oldmode = umask(S_IRGRP|S_IWGRP|S_IXGRP|
 			S_IROTH|S_IWOTH|S_IXOTH);
 	fd = mkostemp(n, O_RDWR|O_CREAT|O_EXCL|O_CLOEXEC);
+	if (fd < 0)
+		fd = -errno;
 	umask(oldmode);
 
 	if (fd >= 0 && name)
@@ -850,7 +952,7 @@ int mnt_open_uniq_filename(const char *filename, char **name)
 	else
 		free(n);
 
-	return fd < 0 ? -errno : fd;
+	return fd;
 }
 
 /**
@@ -860,7 +962,11 @@ int mnt_open_uniq_filename(const char *filename, char **name)
  * This function finds the mountpoint that a given path resides in. @path
  * should be canonicalized. The returned pointer should be freed by the caller.
  *
- * Returns: allocated string with target of the mounted device or NULL on error
+ * WARNING: the function compares st_dev of the @path elements. This traditional
+ * way maybe be insufficient on filesystems like Linux "overlay". See also
+ * mnt_table_find_target().
+ *
+ * Returns: allocated string with the target of the mounted device or NULL on error
  */
 char *mnt_get_mountpoint(const char *path)
 {
@@ -868,7 +974,8 @@ char *mnt_get_mountpoint(const char *path)
 	struct stat st;
 	dev_t dir, base;
 
-	assert(path);
+	if (!path)
+		return NULL;
 
 	mnt = strdup(path);
 	if (!mnt)
@@ -876,7 +983,7 @@ char *mnt_get_mountpoint(const char *path)
 	if (*mnt == '/' && *(mnt + 1) == '\0')
 		goto done;
 
-	if (stat(mnt, &st))
+	if (mnt_stat_mountpoint(mnt, &st))
 		goto err;
 	base = st.st_dev;
 
@@ -885,7 +992,7 @@ char *mnt_get_mountpoint(const char *path)
 
 		if (!p)
 			break;
-		if (stat(*mnt ? mnt : "/", &st))
+		if (mnt_stat_mountpoint(*mnt ? mnt : "/", &st))
 			goto err;
 		dir = st.st_dev;
 		if (dir != base) {
@@ -898,58 +1005,40 @@ char *mnt_get_mountpoint(const char *path)
 
 	memcpy(mnt, "/", 2);
 done:
-	DBG(UTILS, mnt_debug("%s mountpoint is %s", path, mnt));
+	DBG(UTILS, ul_debug("%s mountpoint is %s", path, mnt));
 	return mnt;
 err:
 	free(mnt);
 	return NULL;
 }
 
-char *mnt_get_fs_root(const char *path, const char *mnt)
-{
-	char *m = (char *) mnt, *res;
-	const char *p;
-	size_t sz;
-
-	if (!m)
-		m = mnt_get_mountpoint(path);
-	if (!m)
-		return NULL;
-
-	sz = strlen(m);
-	p = sz > 1 ? path + sz : path;
-
-	if (m != mnt)
-		free(m);
-
-	res = *p ? strdup(p) : strdup("/");
-	DBG(UTILS, mnt_debug("%s fs-root is %s", path, res));
-	return res;
-}
-
 /*
  * Search for @name kernel command parametr.
  *
- * Returns newly allocated string with parameter argument if the @name is
+ * Returns newly allocated string with a parameter argument if the @name is
  * specified as "name=" or returns pointer to @name or returns NULL if not
- * found.
+ * found.  If it is specified more than once, we grab the last copy.
  *
  * For example cmdline: "aaa bbb=BBB ccc"
  *
  *	@name is "aaa"	--returns--> "aaa" (pointer to @name)
  *	@name is "bbb=" --returns--> "BBB" (allocated)
  *	@name is "foo"  --returns--> NULL
+ *
+ * Note: It is not really feasible to parse the command line exactly the same
+ * as the kernel does since we don't know which options are valid.  We can use
+ * the -- marker though and not walk past that.
  */
 char *mnt_get_kernel_cmdline_option(const char *name)
 {
 	FILE *f;
 	size_t len;
 	int val = 0;
-	char *p, *res = NULL;
+	char *p, *res = NULL, *mem = NULL;
 	char buf[BUFSIZ];	/* see kernel include/asm-generic/setup.h: COMMAND_LINE_SIZE */
 	const char *path = _PATH_PROC_CMDLINE;
 
-	if (!name)
+	if (!name || !name[0])
 		return NULL;
 
 #ifdef TEST_PROGRAM
@@ -967,75 +1056,134 @@ char *mnt_get_kernel_cmdline_option(const char *name)
 	if (!p || !*p || *p == '\n')
 		return NULL;
 
-	len = strlen(buf);
-	*(buf + len - 1) = '\0';	/* remove last '\n' */
+	p = strstr(p, " -- ");
+	if (p) {
+		/* no more kernel args after this */
+		*p = '\0';
+	} else {
+		len = strlen(buf);
+		buf[len - 1] = '\0';	/* remove last '\n' */
+	}
 
 	len = strlen(name);
-	if (len && *(name + len - 1) == '=')
+	if (name[len - 1] == '=')
 		val = 1;
 
-	for ( ; p && *p; p++) {
+	for (p = buf; p && *p; p++) {
 		if (!(p = strstr(p, name)))
 			break;			/* not found the option */
 		if (p != buf && !isblank(*(p - 1)))
 			continue;		/* no space before the option */
 		if (!val && *(p + len) != '\0' && !isblank(*(p + len)))
-			continue;		/* no space behind the option */
+			continue;		/* no space after the option */
 		if (val) {
 			char *v = p + len;
+			int end;
 
 			while (*p && !isblank(*p))	/* jump to the end of the argument */
 				p++;
+			end = (*p == '\0');
 			*p = '\0';
-			res = strdup(v);
-			break;
+			free(mem);
+			res = mem = strdup(v);
+			if (end)
+				break;
 		} else
 			res = (char *) name;	/* option without '=' */
-		break;
+		/* don't break -- keep scanning for more options */
 	}
 
 	return res;
 }
 
-int mkdir_p(const char *path, mode_t mode)
+/*
+ * Converts @devno to the real device name if devno major number is greater
+ * than zero, otherwise use root= kernel cmdline option to get device name.
+ *
+ * The function uses /sys to convert devno to device name.
+ *
+ * Returns: 0 = success, 1 = not found, <0 = error
+ */
+int mnt_guess_system_root(dev_t devno, struct libmnt_cache *cache, char **path)
 {
-	char *p, *dir;
-	int rc = 0;
+	char buf[PATH_MAX];
+	char *dev = NULL, *spec = NULL;
+	unsigned int x, y;
+	int allocated = 0;
 
-	if (!path || !*path)
-		return -EINVAL;
+	assert(path);
 
-	dir = p = strdup(path);
-	if (!dir)
-		return -ENOMEM;
+	DBG(UTILS, ul_debug("guessing system root [devno %u:%u]", major(devno), minor(devno)));
 
-	if (*p == '/')
-		p++;
-
-	while (p && *p) {
-		char *e = strchr(p, '/');
-		if (e)
-			*e = '\0';
-		if (*p) {
-			rc = mkdir(dir, mode);
-			if (rc && errno != EEXIST)
-				break;
-			rc = 0;
+	/* The pseudo-fs, net-fs or btrfs devno is useless, otherwise it
+	 * usually matches with the source device, let's try to use it.
+	 */
+	if (major(devno) > 0) {
+		dev = sysfs_devno_to_devpath(devno, buf, sizeof(buf));
+		if (dev) {
+			DBG(UTILS, ul_debug("  devno converted to %s", dev));
+			goto done;
 		}
-		if (!e)
-			break;
-		*e = '/';
-		p = e + 1;
 	}
 
-	DBG(UTILS, mnt_debug("%s mkdir %s", path, rc ? "FAILED" : "SUCCESS"));
+	/* Let's try to use root= kernel command line option
+	 */
+	spec = mnt_get_kernel_cmdline_option("root=");
+	if (!spec)
+		goto done;
 
-	free(dir);
-	return rc;
+	/* maj:min notation */
+	if (sscanf(spec, "%u:%u", &x, &y) == 2) {
+		dev = sysfs_devno_to_devpath(makedev(x, y), buf, sizeof(buf));
+		if (dev) {
+			DBG(UTILS, ul_debug("  root=%s converted to %s", spec, dev));
+			goto done;
+		}
+
+	/* hexhex notation */
+	} else if (isxdigit_string(spec)) {
+		char *end = NULL;
+		uint32_t n;
+
+		errno = 0;
+		n = strtoul(spec, &end, 16);
+
+		if (errno || spec == end || (end && *end))
+			DBG(UTILS, ul_debug("  failed to parse root='%s'", spec));
+		else {
+			/* kernel new_decode_dev() */
+			x = (n & 0xfff00) >> 8;
+			y = (n & 0xff) | ((n >> 12) & 0xfff00);
+			dev = sysfs_devno_to_devpath(makedev(x, y), buf, sizeof(buf));
+			if (dev) {
+				DBG(UTILS, ul_debug("  root=%s converted to %s", spec, dev));
+				goto done;
+			}
+		}
+
+	/* devname or PARTUUID= etc. */
+	} else {
+		DBG(UTILS, ul_debug("  converting root='%s'", spec));
+
+		dev = mnt_resolve_spec(spec, cache);
+		if (dev && !cache)
+			allocated = 1;
+	}
+done:
+	free(spec);
+	if (dev) {
+		*path = allocated ? dev : strdup(dev);
+		if (!path)
+			return -ENOMEM;
+		return 0;
+	}
+
+	return 1;
 }
 
+
 #ifdef TEST_PROGRAM
-int test_match_fstype(struct libmnt_test *ts, int argc, char *argv[])
+static int test_match_fstype(struct libmnt_test *ts, int argc, char *argv[])
 {
 	char *type = argv[1];
 	char *pattern = argv[2];
@@ -1044,7 +1192,7 @@ int test_match_fstype(struct libmnt_test *ts, int argc, char *argv[])
 	return 0;
 }
 
-int test_match_options(struct libmnt_test *ts, int argc, char *argv[])
+static int test_match_options(struct libmnt_test *ts, int argc, char *argv[])
 {
 	char *optstr = argv[1];
 	char *pattern = argv[2];
@@ -1053,7 +1201,7 @@ int test_match_options(struct libmnt_test *ts, int argc, char *argv[])
 	return 0;
 }
 
-int test_startswith(struct libmnt_test *ts, int argc, char *argv[])
+static int test_startswith(struct libmnt_test *ts, int argc, char *argv[])
 {
 	char *optstr = argv[1];
 	char *pattern = argv[2];
@@ -1062,7 +1210,7 @@ int test_startswith(struct libmnt_test *ts, int argc, char *argv[])
 	return 0;
 }
 
-int test_endswith(struct libmnt_test *ts, int argc, char *argv[])
+static int test_endswith(struct libmnt_test *ts, int argc, char *argv[])
 {
 	char *optstr = argv[1];
 	char *pattern = argv[2];
@@ -1071,7 +1219,19 @@ int test_endswith(struct libmnt_test *ts, int argc, char *argv[])
 	return 0;
 }
 
-int test_mountpoint(struct libmnt_test *ts, int argc, char *argv[])
+static int test_appendstr(struct libmnt_test *ts, int argc, char *argv[])
+{
+	char *str = strdup(argv[1]);
+	const char *ap = argv[2];
+
+	append_string(&str, ap);
+	printf("new string: '%s'\n", str);
+
+	free(str);
+	return 0;
+}
+
+static int test_mountpoint(struct libmnt_test *ts, int argc, char *argv[])
 {
 	char *path = canonicalize_path(argv[1]),
 	     *mnt = path ? mnt_get_mountpoint(path) :  NULL;
@@ -1082,18 +1242,7 @@ int test_mountpoint(struct libmnt_test *ts, int argc, char *argv[])
 	return 0;
 }
 
-int test_fsroot(struct libmnt_test *ts, int argc, char *argv[])
-{
-	char *path = canonicalize_path(argv[1]),
-	     *mnt = path ? mnt_get_fs_root(path, NULL) : NULL;
-
-	printf("%s: %s\n", argv[1], mnt ? : "unknown");
-	free(mnt);
-	free(path);
-	return 0;
-}
-
-int test_filesystems(struct libmnt_test *ts, int argc, char *argv[])
+static int test_filesystems(struct libmnt_test *ts, int argc, char *argv[])
 {
 	char **filesystems = NULL;
 	int rc;
@@ -1108,7 +1257,7 @@ int test_filesystems(struct libmnt_test *ts, int argc, char *argv[])
 	return rc;
 }
 
-int test_chdir(struct libmnt_test *ts, int argc, char *argv[])
+static int test_chdir(struct libmnt_test *ts, int argc, char *argv[])
 {
 	int rc;
 	char *path = canonicalize_path(argv[1]),
@@ -1127,7 +1276,7 @@ int test_chdir(struct libmnt_test *ts, int argc, char *argv[])
 	return rc;
 }
 
-int test_kernel_cmdline(struct libmnt_test *ts, int argc, char *argv[])
+static int test_kernel_cmdline(struct libmnt_test *ts, int argc, char *argv[])
 {
 	char *name = argv[1];
 	char *res;
@@ -1145,7 +1294,34 @@ int test_kernel_cmdline(struct libmnt_test *ts, int argc, char *argv[])
 	return 0;
 }
 
-int test_mkdir(struct libmnt_test *ts, int argc, char *argv[])
+
+static int test_guess_root(struct libmnt_test *ts, int argc, char *argv[])
+{
+	int rc;
+	char *real;
+	dev_t devno = 0;
+
+	if (argc) {
+		unsigned int x, y;
+
+		if (sscanf(argv[1], "%u:%u", &x, &y) != 2)
+			return -EINVAL;
+		devno = makedev(x, y);
+	}
+
+	rc = mnt_guess_system_root(devno, NULL, &real);
+	if (rc < 0)
+		return rc;
+	if (rc == 1)
+		fputs("not found\n", stdout);
+	else {
+		printf("%s\n", real);
+		free(real);
+	}
+	return 0;
+}
+
+static int test_mkdir(struct libmnt_test *ts, int argc, char *argv[])
 {
 	int rc;
 
@@ -1157,6 +1333,21 @@ int test_mkdir(struct libmnt_test *ts, int argc, char *argv[])
 	return rc;
 }
 
+static int test_statfs_type(struct libmnt_test *ts, int argc, char *argv[])
+{
+	struct statfs vfs;
+	int rc;
+
+	rc = statfs(argv[1], &vfs);
+	if (rc)
+		printf("%s: statfs failed: %m\n", argv[1]);
+	else
+		printf("%-30s: statfs type: %-12s [0x%lx]\n", argv[1],
+				mnt_statfs_get_fstype(&vfs),
+				(long) vfs.f_type);
+	return rc;
+}
+
 
 int main(int argc, char *argv[])
 {
@@ -1166,11 +1357,13 @@ int main(int argc, char *argv[])
 	{ "--filesystems",   test_filesystems,	   "[<pattern>] list /{etc,proc}/filesystems" },
 	{ "--starts-with",   test_startswith,      "<string> <prefix>" },
 	{ "--ends-with",     test_endswith,        "<string> <prefix>" },
+	{ "--append-string", test_appendstr,       "<string> <appendix>" },
 	{ "--mountpoint",    test_mountpoint,      "<path>" },
-	{ "--fs-root",       test_fsroot,          "<path>" },
 	{ "--cd-parent",     test_chdir,           "<path>" },
 	{ "--kernel-cmdline",test_kernel_cmdline,  "<option> | <option>=" },
+	{ "--guess-root",    test_guess_root,      "[<maj:min>]" },
 	{ "--mkdir",         test_mkdir,           "<path>" },
+	{ "--statfs-type",   test_statfs_type,     "<path>" },
 
 	{ NULL }
 	};
