@@ -43,9 +43,8 @@ static int table_parser_errcb(struct libmnt_table *tb __attribute__((__unused__)
 			const char *filename, int line)
 {
 	if (filename)
-		warnx(_("%s: parse error: ignore entry at line %d."),
-							filename, line);
-	return 0;
+		warnx(_("%s: parse error at line %d -- ignored"), filename, line);
+	return 1;
 }
 
 
@@ -78,6 +77,9 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 		" %1$s [options] <source> | <directory>\n"),
 		program_invocation_short_name);
 
+	fputs(USAGE_SEPARATOR, out);
+	fputs(_("Unmount filesystems.\n"), out);
+
 	fputs(USAGE_OPTIONS, out);
 	fputs(_(" -a, --all               unmount all filesystems\n"), out);
 	fputs(_(" -A, --all-targets       unmount all mountpoints for the given device in the\n"
@@ -435,7 +437,9 @@ static int umount_alltargets(struct libmnt_context *cxt, const char *spec, int r
 		return mk_exit_code(cxt, rc);		/* error */
 
 	if (!mnt_fs_get_srcpath(fs) || !mnt_fs_get_devno(fs))
-		err(MOUNT_EX_USAGE, _("%s: failed to determine source"), spec);
+		errx(MOUNT_EX_USAGE, _("%s: failed to determine source "
+				"(--all-targets is unsupported on systems with "
+				"regular mtab file)."), spec);
 
 	itr = mnt_new_iter(MNT_ITER_BACKWARD);
 	if (!itr)
@@ -443,8 +447,10 @@ static int umount_alltargets(struct libmnt_context *cxt, const char *spec, int r
 
 	/* get on @cxt independent mountinfo */
 	tb = new_mountinfo(cxt);
-	if (!tb)
-		return MOUNT_EX_SOFTWARE;
+	if (!tb) {
+		rc = MOUNT_EX_SOFTWARE;
+		goto done;
+	}
 
 	/* Note that @fs is from mount context and the context will be reseted
 	 * after each umount() call */
@@ -466,6 +472,7 @@ static int umount_alltargets(struct libmnt_context *cxt, const char *spec, int r
 			break;
 	}
 
+done:
 	mnt_free_iter(itr);
 	mnt_unref_table(tb);
 
@@ -644,6 +651,6 @@ int main(int argc, char **argv)
 	}
 
 	mnt_free_context(cxt);
-	return rc;
+	return (rc < 256) ? rc : 255;
 }
 
