@@ -19,13 +19,18 @@
  * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
  * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
  * OTHER DEALINGS IN THE SOFTWARE.
- *
- * less -F and tail -f cause a disk access every five seconds.  This
- * program avoids this problem by waiting for the file size to change.
- * Hence, the file is not accessed, and the access time does not need to be
- * flushed back to disk.  This is sort of a "stealth" tail.
  */
 
+/*
+ * This command is deprecated.  The utility is in maintenance mode,
+ * meaning we keep them in source tree for backward compatibility
+ * only.  Do not waste time making this command better, unless the
+ * fix is about security or other very critical issue.
+ *
+ * See Documentation/deprecated.txt for more information.
+ */
+
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -36,6 +41,9 @@
 #include <ctype.h>
 #include <errno.h>
 #include <getopt.h>
+#include <sys/mman.h>
+#include <limits.h>
+
 #ifdef HAVE_INOTIFY_INIT
 #include <sys/inotify.h>
 #endif
@@ -48,45 +56,41 @@
 
 #define DEFAULT_LINES  10
 
-static void
-tailf(const char *filename, int lines)
+/* st->st_size has to be greater than zero and smaller or equal to SIZE_MAX! */
+static void tailf(const char *filename, size_t lines, struct stat *st)
 {
-	char *buf, *p;
-	int  head = 0;
-	int  tail = 0;
-	FILE *str;
-	int  i;
+	int fd;
+	size_t i;
+	char *data;
 
-	if (!(str = fopen(filename, "r")))
+	fd = open(filename, O_RDONLY);
+	if (fd < 0)
 		err(EXIT_FAILURE, _("cannot open %s"), filename);
+	data = mmap(0, st->st_size, PROT_READ, MAP_SHARED, fd, 0);
+	i = (size_t) st->st_size - 1;
 
-	buf = xmalloc((lines ? lines : 1) * BUFSIZ);
-	p = buf;
-	while (fgets(p, BUFSIZ, str)) {
-		if (++tail >= lines) {
-			tail = 0;
-			head = 1;
+	/* humans do not think last new line in a file should be counted,
+	 * in that case do off by one from counter point of view */
+	if (data[i] == '\n')
+		lines++;
+	while (i) {
+		if (data[i] == '\n') {
+			if (--lines == 0) {
+				i++;
+				break;
+			}
 		}
-		p = buf + (tail * BUFSIZ);
+		i--;
 	}
 
-	if (head) {
-		for (i = tail; i < lines; i++)
-			fputs(buf + (i * BUFSIZ), stdout);
-		for (i = 0; i < tail; i++)
-			fputs(buf + (i * BUFSIZ), stdout);
-	} else {
-		for (i = head; i < tail; i++)
-			fputs(buf + (i * BUFSIZ), stdout);
-	}
+	fwrite(data + i, st->st_size - i, 1, stdout);
 
+	munmap(data, st->st_size);
+	close(fd);
 	fflush(stdout);
-	free(buf);
-	fclose(str);
 }
 
-static void
-roll_file(const char *filename, off_t *size)
+static void roll_file(const char *filename, struct stat *old)
 {
 	char buf[BUFSIZ];
 	int fd;
@@ -100,12 +104,12 @@ roll_file(const char *filename, off_t *size)
 	if (fstat(fd, &st) == -1)
 		err(EXIT_FAILURE, _("stat of %s failed"), filename);
 
-	if (st.st_size == *size) {
+	if (st.st_size == old->st_size) {
 		close(fd);
 		return;
 	}
 
-	if (lseek(fd, *size, SEEK_SET) != (off_t)-1) {
+	if (lseek(fd, old->st_size, SEEK_SET) != (off_t)-1) {
 		ssize_t rc, wc;
 
 		while ((rc = read(fd, buf, sizeof(buf))) > 0) {
@@ -123,16 +127,15 @@ roll_file(const char *filename, off_t *size)
 	 * avoids data duplication. If we read nothing or hit an error, reset
 	 * to the reported size, this handles truncated files.
 	 */
-	*size = (pos != -1 && pos != *size) ? pos : st.st_size;
+	old->st_size = (pos != -1 && pos != old->st_size) ? pos : st.st_size;
 
 	close(fd);
 }
 
-static void
-watch_file(const char *filename, off_t *size)
+static void watch_file(const char *filename, struct stat *old)
 {
 	do {
-		roll_file(filename, size);
+		roll_file(filename, old);
 		xusleep(250000);
 	} while(1);
 }
@@ -143,8 +146,7 @@ watch_file(const char *filename, off_t *size)
 #define EVENTS		(IN_MODIFY|IN_DELETE_SELF|IN_MOVE_SELF|IN_UNMOUNT)
 #define NEVENTS		4
 
-static int
-watch_file_inotify(const char *filename, off_t *size)
+static int watch_file_inotify(const char *filename, struct stat *old)
 {
 	char buf[ NEVENTS * sizeof(struct inotify_event) ];
 	int fd, ffd, e;
@@ -176,7 +178,7 @@ watch_file_inotify(const char *filename, off_t *size)
 			struct inotify_event *ev = (struct inotify_event *) &buf[e];
 
 			if (ev->mask & IN_MODIFY)
-				roll_file(filename, size);
+				roll_file(filename, old);
 			else {
 				close(ffd);
 				ffd = -1;
@@ -207,21 +209,22 @@ static void __attribute__ ((__noreturn__)) usage(FILE *out)
 	fputs(USAGE_HELP, out);
 	fputs(USAGE_VERSION, out);
 	fprintf(out, USAGE_MAN_TAIL("tailf(1)"));
+	fputs(_("Warning: use of 'tailf' is deprecated, use 'tail -f' instead.\n"), out);
 
 	exit(out == stderr ? EXIT_FAILURE : EXIT_SUCCESS);
 }
 
 /* parses -N option */
-static long old_style_option(int *argc, char **argv)
+static long old_style_option(int *argc, char **argv, size_t *lines)
 {
-	int i = 1, nargs = *argc;
-	long lines = -1;
+	int i = 1, nargs = *argc, ret = 0;
 
 	while(i < nargs) {
 		if (argv[i][0] == '-' && isdigit(argv[i][1])) {
-			lines = strtol_or_err(argv[i] + 1,
+			*lines = strtoul_or_err(argv[i] + 1,
 					_("failed to parse number of lines"));
 			nargs--;
+			ret = 1;
 			if (nargs - i)
 				memmove(argv + i, argv + i + 1,
 						sizeof(char *) * (nargs - i));
@@ -229,16 +232,15 @@ static long old_style_option(int *argc, char **argv)
 			i++;
 	}
 	*argc = nargs;
-	return lines;
+	return ret;
 }
 
 int main(int argc, char **argv)
 {
 	const char *filename;
-	long lines;
+	size_t lines;
 	int ch;
 	struct stat st;
-	off_t size = 0;
 
 	static const struct option longopts[] = {
 		{ "lines",   required_argument, 0, 'n' },
@@ -252,15 +254,14 @@ int main(int argc, char **argv)
 	textdomain(PACKAGE);
 	atexit(close_stdout);
 
-	lines = old_style_option(&argc, argv);
-	if (lines < 0)
+	if (!old_style_option(&argc, argv, &lines))
 		lines = DEFAULT_LINES;
 
 	while ((ch = getopt_long(argc, argv, "n:N:Vh", longopts, NULL)) != -1)
-		switch((char)ch) {
+		switch ((char)ch) {
 		case 'n':
 		case 'N':
-			lines = strtol_or_err(optarg,
+			lines = strtoul_or_err(optarg,
 					_("failed to parse number of lines"));
 			break;
 		case 'V':
@@ -279,14 +280,17 @@ int main(int argc, char **argv)
 
 	if (stat(filename, &st) != 0)
 		err(EXIT_FAILURE, _("stat of %s failed"), filename);
+	if (!S_ISREG(st.st_mode))
+		errx(EXIT_FAILURE, _("%s: is not a file"), filename);
 
-	size = st.st_size;;
-	tailf(filename, lines);
+	/* mmap is based on size_t */
+	if (st.st_size > 0 && (uintmax_t) st.st_size <= (uintmax_t) SIZE_MAX)
+		tailf(filename, lines, &st);
 
 #ifdef HAVE_INOTIFY_INIT
-	if (!watch_file_inotify(filename, &size))
+	if (!watch_file_inotify(filename, &st))
 #endif
-		watch_file(filename, &size);
+		watch_file(filename, &st);
 
 	return EXIT_SUCCESS;
 }
